/*
 * FogLAMP interface to the bucket service
 *
 * Copyright (c) 2023 Dianomic Systems
 *
 * Released under the Apache 2.0 Licence
 *
 * Author: Mark Riddoch
 */
#include <bucket_client.h>
#include <service_handler.h>
#include <rapidjson/document.h>

using namespace std;
using namespace rapidjson;

BucketClient *BucketClient::instance = 0;

/**
 * Return the singleton bucket client instance
 *
 * @return BucketClient*	Singleton instance or NULL on failure
 */
BucketClient *
BucketClient::getInstance()
{
	if (!instance)
		instance = new BucketClient();
	return instance;
}

/**
 * Constructor for the bucket service client
 */
BucketClient::BucketClient() : m_managementClient(NULL)
{
	m_managementClient = ServiceAuthHandler::getMgmtClient();
	m_logger = Logger::getLogger();
}

/**
 * Destroy the bucket client singleton
 */
BucketClient::~BucketClient()
{
	lock_guard<mutex> guard(m_mutex);
	for (auto& s : m_servers)
		delete s;
	m_servers.clear();
	instance = NULL;
}

/**
 * Get the list of bucket servers running in the instance
 *
 * @return bool		True if we found any servers
 */
bool BucketClient::getServers()
{
	lock_guard<mutex> guard(m_mutex);
	if (m_servers.size() > 0)
	{
		for (auto& s : m_servers)
			delete s;
		m_servers.clear();
	}
	if (!m_managementClient)
	{
		m_managementClient = ServiceAuthHandler::getMgmtClient();
	}
	if (m_managementClient)
	{
		return m_managementClient->getServices(m_servers, BUCKET_SERVICE_TYPE);
	}
	else
	{
		m_logger->fatal("Unable to find management interface, buckets can not be retrieved");
	}
	return false;
}

/**
 * Find a matching bucket for the supplied configuration category
 * item
 *
 * @param config	The configuration category
 * @param name		The name of the configuration item
 * @param matches	Vector of bucket matches to return
 * @return int		The number of matching buckets or -1 on error
 */
int BucketClient::matchBucket(const ConfigCategory& config, const string& name, vector<BucketMatch>& matches)
{
	if (!config.itemExists(name))
	{
		 m_logger->error("The configuration category does not contain an item called '%s'", name.c_str());
		 return -1;
	}

	if (m_servers.size() == 0)
	{
		getServers();
	}
	if (m_servers.size() == 0)
	{
		return -1;
	}
	string value = config.getValue(name);
	string type = config.getType(name);
	if (type.compare(BUCKET_ITEM_TYPE) != 0)
	{
		m_logger->error("The configuration category item %s is not of the correct type to reference a storage bucket", name.c_str());
		return -1;
	}

	// Build the request payload to send to the bucket server
	string payload = " { \"attributes\" : ";
	payload += value;
	payload += "}";

	ServiceRecord *service = m_servers[0];
	string url = "/bucket/match";

	string message, status;
	string address = service->getAddress() + ":" + to_string(service->getPort());
	m_logger->debug("Call %s%s with payload: %s", address.c_str(), url.c_str(), payload.c_str());
	if (service->getProtocol().compare("http"))
	{
		SimpleWeb::Client<SimpleWeb::HTTP> http(address);
		try {
			auto res = http.request("PUT", url, payload);
			message = res->content.string();
			status = res->status_code;
		} catch (exception& e) {
			m_logger->error("Failed to call bucket storage service match api: %s", e.what());
			return -1;
		}
	}
	else if (service->getProtocol().compare("https"))
	{
		SimpleWeb::Client<SimpleWeb::HTTP> https(address);
		try {
			auto res = https.request("PUT", url, payload);
			message = res->content.string();
			status = res->status_code;
		} catch (exception& e) {
			m_logger->error("Failed to call bucket storage service match api: %s", e.what());
			return -1;
		}
	}

	if (status[0] != '2')
	{
		m_logger->error("Failed to execute match against bucket service: %s, %s",
				status.c_str(), message.c_str());
		return -1;
	}

	m_logger->debug("Response is: %s", message.c_str());
	Document doc;
	doc.Parse(message.c_str());
	if (doc.HasMember("matches"))
	{
		const Value& match = doc["matches"];
		if (! match.IsArray())
		{
			m_logger->error("The returned matches from the bucket storage service is not an array");
			return -1;
		}
		if (match.Empty())
		{
			// No Matches
			m_logger->debug("No matching bucket for the attributes %s", value.c_str());
			return 0;
		}
		for (auto& m : match.GetArray())
		{
			if (m.IsObject())
			{
				int id = -1;
				string file;
				if (m.HasMember("bucket") && m["bucket"].IsInt64())
				{
					id = m["bucket"].GetInt64();
				}
				if (m.HasMember("file") && m["file"].IsString())
				{
					file = m["file"].GetString();
				}
				if (id == -1 && file.empty())
				{
				}
				else
				{
					matches.push_back(BucketMatch(id, file));
				}
			}
		}
		return matches.size();
	}
	
	return -1;
}
